from .daily import Daily
from .hourly import Hourly

# from importlib import metadata
# __version__ = metadata.version(__package__.replace('.', '-') or __name__.replace('.', '-'))
# # __version__ = metadata.version(__package__ or __name__)
# # __version__ = metadata.version('openet-refet-gee')

